#include<string>
#include"STATUS.hpp"
/* By A.N. Yzelman, http://www.math.uu.nl/people/yzelman
 * Software is released under GPLv3, see LICENSE.txt
 */


/*! \mainpage Cache Simulator -- perfect cache simulator
 *
 * The cache simulator uses the CACHE, CS_ELEMENT, CS_ARRAY, CS_MATRIX, CS_CRS and
 * CS_ICRS classes to simulate cache behaviour in order to obtain cache performance
 * statistics.
 *
 * This is <em>not</em> a trace-driven cache simulator. Applications simply
 * use the classes provided by this library instead of normal datatypes.
 *
 * This library is developed as a means to obtain cache hit/miss
 * statistics in the area of sparse matrix operations, using our own idealised
 * cache model:
 *
 * -The cache is k-way set-associative cache.
 *
 * -There is only one cache in between the CPU and main memory;
 *  i.e., we assume a single-level cache architecture.
 *
 * -The cache is always full; that is, cache lines always store valid data.
 *  Hence a cache miss always results in a cache eviction.
 *
 * -The replacement policy when choosing between $k$ cache lines is the Least Recently
 *  Used (LRU) policy.
 *
 * -Reading in n cache lines from main memory to the cache takes O(l_m+nb_m) time,
 *  where l_m,b_m in (0,infinity) constant.
 *
 * -Reading in n cache lines from cache to CPU takes O(l_c+nb_c) time, where l_c,b_c
 *  in (0,infinity) constant.
 *
 * -Analogously: writing n cache lines from CPU to cache takes O(l_c+nb_c) time,
 *  whereas writing from cache to main memory takes O(l_m+nb_m) time.
 *
 *  The current framework only keeps statistics on the number of cache hits and misses.
 */

#ifndef _H_CS_CACHE
#define _H_CS_CACHE

/** The cache simulator main class. */
class CACHE {
   public:

	/** Returns the singleton cache. If not already initialised, the singleton will be set to the default cache with parameters:
	    l=2^{26}
	    s=2^{32}
	    k=2^4
	    (Corresponding to an Intel Core 2 4MB L2 cache)
	*/
	static CACHE* getInstance();

	/** Creates an instance with specific cache parameters. */
	static void createInstance( unsigned long int l, unsigned long int s, unsigned long int k );

	/** Deletes the current cache instance and creates a new one with the given cache parameters. */
	static void recreateInstance( unsigned long int l, unsigned long int s, unsigned long int k );

	/** Resets all counters. */
	void clearStatistics();

	/** Returns the current cache status. */
	const STATUS &getStatus();
	
	/** Main constructor. */
	CACHE( unsigned long int l, unsigned long int s, unsigned long int k );
	
	/** Detects a cache hit or miss. Does not update statistics! */
	bool hitOrMiss( void* pointer, unsigned long int size );
	
	/** Simulates a read. */
	void access( void* pointer, const unsigned long int size, unsigned long int &FAC );

	/** Prints cache properties to output string. */
	std::string properties();

	/** Clears all cache contents. */
	void clear();
	
	/* Simulates a write. */
	//void write( void* pointer, unsigned long int size );
	
};

#endif

